import time
from pathlib import Path

import fast_align_audio
import numpy as np
import pytest
import soundfile
import scipy.signal
import fast_mp3_augment

TEST_FIXTURES_PATH = Path(__file__).resolve().parent.parent / "test_fixtures"


def get_chirp_test(sample_rate, duration):
    """Create a `duration` seconds chirp from 0 Hz to Nyquist frequency"""
    n = np.arange(0, duration, 1 / sample_rate)
    samples = scipy.signal.chirp(n, 0, duration, sample_rate // 2, method="linear")
    return samples.astype(np.float32)


def test_mono_1d():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "p286_011.wav", dtype=np.float32
    )
    assert audio.ndim == 1
    augmented_audio = fast_mp3_augment.compress_roundtrip(
        audio, sample_rate, bitrate_kbps=64
    )
    assert augmented_audio.shape == audio.shape
    assert augmented_audio.dtype == augmented_audio.dtype

    offset, mse = fast_align_audio.find_best_alignment_offset(
        reference_signal=audio,
        delayed_signal=augmented_audio,
        max_offset_samples=3000,
    )
    assert mse <= 5e-5
    assert offset == 0


def test_mono_1d_preserve_delay():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "p286_011.wav", dtype=np.float32
    )
    assert audio.ndim == 1
    augmented_audio = fast_mp3_augment.compress_roundtrip(
        audio,
        sample_rate,
        bitrate_kbps=64,
        preserve_delay=True,
    )

    assert augmented_audio.shape[-1] > audio.shape[-1]
    assert augmented_audio.dtype == augmented_audio.dtype

    offset, mse = fast_align_audio.find_best_alignment_offset(
        reference_signal=audio,
        delayed_signal=augmented_audio,
        max_offset_samples=3000,
    )
    assert offset > 0

    padding = augmented_audio.shape[-1] - audio.shape[-1] - offset
    assert padding > 0


def test_mono_2d():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "p286_011.wav", dtype=np.float32
    )
    audio = np.expand_dims(audio, 0)
    assert audio.ndim == 2
    augmented_audio = fast_mp3_augment.compress_roundtrip(
        audio, sample_rate, bitrate_kbps=64
    )
    assert augmented_audio.shape == audio.shape
    assert augmented_audio.dtype == augmented_audio.dtype


def test_stereo():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "perfect-alley1.ogg", dtype=np.float32
    )
    audio = np.ascontiguousarray(audio.T)
    augmented_audio = fast_mp3_augment.compress_roundtrip(
        audio, sample_rate, bitrate_kbps=64
    )

    assert augmented_audio.shape == audio.shape
    assert augmented_audio.dtype == augmented_audio.dtype

    offset, mse = fast_align_audio.find_best_alignment_offset(
        reference_signal=audio[0],
        delayed_signal=augmented_audio[0],
        max_offset_samples=3000,
    )
    assert mse <= 5e-5
    assert offset == 0


def test_stereo_preserve_delay():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "perfect-alley1.ogg", dtype=np.float32
    )
    audio = np.ascontiguousarray(audio.T)
    augmented_audio = fast_mp3_augment.compress_roundtrip(
        audio, sample_rate, bitrate_kbps=64, preserve_delay=True
    )

    assert augmented_audio.shape[0] == audio.shape[0]
    assert augmented_audio.shape[1] > audio.shape[1]
    assert augmented_audio.dtype == augmented_audio.dtype

    offset, mse = fast_align_audio.find_best_alignment_offset(
        reference_signal=audio[0],
        delayed_signal=augmented_audio[0],
        max_offset_samples=3000,
    )
    assert offset > 0

    padding = augmented_audio.shape[1] - audio.shape[1] - offset
    assert padding > 0


def test_stereo_wrong_dimension_order():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "perfect-alley1.ogg", dtype=np.float32
    )
    assert audio.shape[1] == 2
    with pytest.raises(fast_mp3_augment.WrongMultichannelAudioShape):
        fast_mp3_augment.compress_roundtrip(audio, sample_rate, bitrate_kbps=64)


def test_3_channels():
    sample_rate = 48000
    audio = np.random.uniform(-1, 1, (3, 48000)).astype("float32")
    with pytest.raises(fast_mp3_augment.UnsupportedNumberOfChannels):
        fast_mp3_augment.compress_roundtrip(audio, sample_rate, bitrate_kbps=64)


def test_too_short():
    sample_rate = 44100
    audio = np.random.uniform(-1, 1, (1, 100)).astype("float32")
    with pytest.raises(fast_mp3_augment.TooShortAudio):
        fast_mp3_augment.compress_roundtrip(audio, sample_rate, bitrate_kbps=64)


def test_transposed_but_not_contiguous_audio():
    audio, sample_rate = soundfile.read(
        TEST_FIXTURES_PATH / "perfect-alley1.ogg", dtype=np.float32
    )
    transposed_non_contiguous_array = audio.T
    with pytest.raises(fast_mp3_augment.NonContiguousAudio):
        fast_mp3_augment.compress_roundtrip(
            transposed_non_contiguous_array, sample_rate, bitrate_kbps=64
        )


def test_supported_sample_rates():
    sample_rates = (8000, 11025, 12000, 16000, 22050, 24000, 32000, 44100, 48000)
    duration_s = 2

    for sample_rate in sample_rates:
        for bitrate_kbps in [
            8,
            16,
            24,
            32,
            40,
            48,
            56,
            64,
            80,
            96,
            112,
            128,
            144,
            160,
            192,
            224,
            256,
            320,
        ]:
            for num_channels in (1, 2):
                sig = get_chirp_test(sample_rate, duration_s)
                sig = np.ascontiguousarray(sig.reshape((num_channels, -1)))
                augmented_sig = fast_mp3_augment.compress_roundtrip(
                    sig, sample_rate, bitrate_kbps=bitrate_kbps
                )
                assert augmented_sig.shape == sig.shape
                offset, mse = fast_align_audio.find_best_alignment_offset(
                    reference_signal=sig[0],
                    delayed_signal=augmented_sig[0],
                    max_offset_samples=3000,
                )
                assert offset == 0

                corr = fast_align_audio.alignment.fast_autocorr(
                    sig.flatten(), augmented_sig.flatten(), t=0
                )
                assert not np.any(np.isnan(augmented_sig))
                if bitrate_kbps == 96:
                    assert corr > 0.75


def test_unsupported_sample_rate():
    sig = np.zeros(4500, dtype=np.float32)
    with pytest.raises(fast_mp3_augment.UnsupportedSampleRate):
        fast_mp3_augment.compress_roundtrip(sig, sample_rate=1337, bitrate_kbps=64)


def test_supported_bitrates():
    bitrates = {
        8,
        16,
        24,
        32,
        40,
        48,
        56,
        64,
        80,
        96,
        112,
        128,
        144,
        160,
        192,
        224,
        256,
        320,
    }
    duration_s = 2
    sample_rate = 44100
    for num_channels in (1, 2):
        for bitrate_kbps in bitrates:
            sig = 0.5 * np.sin(
                2
                * np.pi
                * 440
                * np.linspace(0, 5, sample_rate * duration_s, dtype=np.float32)
            )
            sig = sig.reshape((num_channels, -1))
            augmented_sig = fast_mp3_augment.compress_roundtrip(
                sig, sample_rate, bitrate_kbps=bitrate_kbps
            )
            assert augmented_sig.shape == sig.shape
            assert not np.any(np.isnan(augmented_sig))


def test_unsupported_bitrate():
    sig = np.zeros(4500, dtype=np.float32)
    with pytest.raises(fast_mp3_augment.UnsupportedBitrate):
        fast_mp3_augment.compress_roundtrip(sig, sample_rate=44100, bitrate_kbps=42)


def test_quality_values():
    quality_values = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    sig = np.random.uniform(-1, 1, 20000).astype("float32")
    execution_times = []
    for quality in quality_values:
        start_t = time.time()
        augmented_sig = fast_mp3_augment.compress_roundtrip(
            sig, sample_rate=44100, bitrate_kbps=64, quality=quality
        )
        execution_time = time.time() - start_t
        execution_times.append(execution_time)
        assert augmented_sig.shape == sig.shape

    # quality = 0 is slower than quality = 9
    assert execution_times[0] > execution_times[-1]


def test_unsupported_quality_value():
    sig = np.zeros(4500, dtype=np.float32)
    with pytest.raises(fast_mp3_augment.UnsupportedQuality):
        fast_mp3_augment.compress_roundtrip(
            sig, sample_rate=44100, bitrate_kbps=64, quality=42
        )



from typing import Final

import numpy as np
from numpy.typing import NDArray
import numpy_minmax

from ._mp3augment import compress_roundtrip as _compress_roundtrip


class WrongMultichannelAudioShape(Exception):
    pass


class UnsupportedNumberOfChannels(Exception):
    pass


class TooShortAudio(Exception):
    pass


class NonContiguousAudio(Exception):
    pass


class UnsupportedSampleRate(Exception):
    pass


class UnsupportedBitrate(Exception):
    pass


class UnsupportedQuality(Exception):
    pass


SUPPORTED_SAMPLE_RATES = {8000, 11025, 12000, 16000, 22050, 24000, 32000, 44100, 48000}
SUPPORTED_BITRATES = {
    8,
    16,
    24,
    32,
    40,
    48,
    56,
    64,
    80,
    96,
    112,
    128,
    144,
    160,
    192,
    224,
    256,
    320,
}
SUPPORTED_QUALITY_VALUES = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}


def compress_roundtrip(
    samples: NDArray[np.float32],
    sample_rate: int,
    bitrate_kbps: int,
    preserve_delay: bool = False,
    quality: int = 7,
) -> NDArray[np.float32]:
    """
    Degrade the given float32 audio by encoding it as MP3 and then decoding it back to a float32 audio signal.

    :param samples: 1D or 2D samples with shape (channels, samples)
    :param sample_rate: Sample rate of the audio in Hertz
    :param bitrate_kbps: Constant bitrate (in kbps) to use when encoding the audio
    :param preserve_delay:
        If False (default), the output length and timing will match the input.
        If True, include LAME encoder delay + filter delay (a few tens of milliseconds) and padding in the output.
        This makes the output
        1) longer than the input
        2) delayed (out of sync) relative to the input
        Normally, it makes sense to set preserve_delay to False, but if you want outputs that include the
        short, almost silent part in the beginning, you here have the option to get that.
    :param quality: An int in range [0, 9].
        0: higher quality audio at the cost of slower processing
        9: fast processing at the cost of lower quality audio
    """
    if sample_rate not in SUPPORTED_SAMPLE_RATES:
        raise UnsupportedSampleRate(
            f"Expected sample_rate to be one of {SUPPORTED_SAMPLE_RATES}, but received {sample_rate}"
        )
    if bitrate_kbps not in SUPPORTED_BITRATES:
        raise UnsupportedBitrate(
            f"Expected bitrate_kbps to be one of {SUPPORTED_BITRATES}, but received {bitrate_kbps}"
        )
    if quality not in SUPPORTED_QUALITY_VALUES:
        raise UnsupportedQuality(
            f"Expected quality to be one of {SUPPORTED_QUALITY_VALUES}, but received {quality}"
        )

    original_ndim = samples.ndim
    if samples.ndim == 2:
        if samples.shape[0] > samples.shape[1]:
            raise WrongMultichannelAudioShape(
                "Multichannel audio must have channels first, not channels"
                " last. In other words, the shape must be (channels, samples),"
                " not (samples, channels). See"
                " https://iver56.github.io/audiomentations/guides/multichannel_audio_array_shapes/"
                " for more info."
            )
        if samples.shape[0] > 2:
            raise UnsupportedNumberOfChannels(
                f"Expected mono or stereo audio, but received {samples.shape[0]} channels, which is unsupported."
            )

        if not samples.flags.c_contiguous:
            raise NonContiguousAudio(
                f"The given NDArray must be C-contiguous. You can make it C-contiguous by using `np.ascontiguousarray(samples)`"
            )

    if samples.shape[-1] < 0.1 * sample_rate:
        raise TooShortAudio(
            f"The input audio {samples.shape[-1]} samples is too short. It needs to be at least 100 ms."
        )

    # Possibly gain down to avoid clipping
    min_amplitude, max_amplitude = numpy_minmax.minmax(samples)
    max_abs_amplitude = max(abs(min_amplitude), abs(max_amplitude))
    post_gain = 1.0
    if max_abs_amplitude > 1.0:
        factor = 1.0 / max_abs_amplitude
        samples = samples * factor
        post_gain = max_abs_amplitude

    result = _compress_roundtrip(
        samples,
        sample_rate,
        bitrate_kbps,
        post_gain=post_gain,
        preserve_delay=preserve_delay,
        quality=quality,
    )
    if result.ndim == 1 and original_ndim == 2:
        result = result.reshape((1, -1))
    return result


__all__: Final = ["compress_roundtrip"]
